/*

    Copyright (C) 2000-2001 Stefan Westerfeld
                            <stefan@space.twc.de>
                       2003 Arnold Krille <arnold@arnoldarts.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    */

#include "main.h"
#include "environmentview.h"
#include "main.moc"
#include "levelmeters.h"
#include "midimanagerview.h"
#include "audiomanager.h"
#include "fftscopeview.h"
#include "mediatypesview.h"
#include "statusview.h"

#include <objectmanager.h>
#include <debug.h>
#include <artsversion.h>
#include <kartsfloatwatch.h>

#include <qlabel.h>
#include <qslider.h>
#include <qpushbutton.h>
#include <kaction.h>
#include <kled.h>
#include <kmessagebox.h>
#include <kdebug.h>
#include <kartsserver.h>

#include <kstdaction.h>

#include "kartswidget.h"

using namespace std;
using namespace Arts;

void FreeVerbView::closeEvent(QCloseEvent *e) {
	e->accept();
	emit closed();
}

FreeVerbView::FreeVerbView(SimpleSoundServer server) : server(server) {
	freeverb = DynamicCast(server.createObject("Arts::Synth_FREEVERB"));
	arts_assert(!freeverb.isNull());
	freeverb.start();

	// put it into the effect chain
	effectID = server.outstack().insertTop(freeverb,"FreeVerb");

	GenericGuiFactory factory;
	Widget gui = factory.createGui(freeverb);

	QBoxLayout * l = new QHBoxLayout( this );
	if(!gui.isNull())
		l->add(new KArtsWidget(gui, this));
	else
		l->add(new QLabel(i18n("No GUI found for this effect."), this));
	l->freeze();
	show();
}

FreeVerbView::~FreeVerbView() {
	// remove effect
	server.outstack().remove(effectID);
}


VControl::VControl(QWidget *parent) : QFrame(parent), scopeView(0), mediaView(0)
	, freeVerbView(0)
	, audioManager(0)
	, artsStatusView(0)
	, midiManagerView(0)
	, environmentView(0)
	, server(KArtsServer().server())
	, inUpdate(false)
{
   	if(server.isNull())
   	{
		KMessageBox::error( 0, i18n("Connection to the soundserver failed - make sure that artsd is really running and that your kdelibs version is not older than kdemultimedia."));
		exit(1);
   	}
	boxLayout = new QHBoxLayout( this );
	//l->setAutoAdd(TRUE);
	// stereoMeter = new StereoLedMeters(this);
	// setMinimumSize(80,300);
	stereoMeter = new PeakLevelMeters(this);
	boxLayout->addWidget(stereoMeter);

	svc = server.outVolume();
	assert(!svc.isNull());

	// 0   => 4.0
	// 200 => 2.0
	// 400 => 1.0
	// 600 => 0.5
	// 800 => 0.25
	// 1000 => 0.125
	// 1200 => 0.0 (forced)

	slider = new QSlider(0,1200,100,400,QSlider::Vertical,this,"volumeslider");
	connect(slider, SIGNAL(valueChanged(int)), this, SLOT(setVolume(int)));
	slider->setTickmarks( QSlider::Right );
	boxLayout->addWidget(slider);

//#if 0
	svcgui = Arts::StereoVolumeControlGui( server.outVolume() );
	svcguiw = new KArtsWidget( svcgui, this );
	boxLayout->addWidget( svcguiw );
	boxLayout->addStretch( 100 );
//#endif

	boxLayout->activate();
	show();
	QTimer *updatescalestimer = new QTimer(this);
	updatescalestimer->start(100);
	connect(updatescalestimer,SIGNAL(timeout()),this,SLOT(updateScales()));

	KArtsFloatWatch *watch = new KArtsFloatWatch(svc, "scaleFactor_changed", this);
	connect(watch, SIGNAL(valueChanged(float)), this, SLOT(updateVolume(float)));
	updateVolume(svc.scaleFactor());
}

VControl::~VControl() {
	if ( scopeView ) showScopeView();
	if ( freeVerbView ) showFreeVerbView();
	if ( audioManager ) showAudioManager();
	if ( mediaView ) showMediaView();
	if ( artsStatusView ) showArtsStatusView();
	if ( midiManagerView ) showMidiManager();
	if ( environmentView ) showEnvironment();
}

void VControl::updateScales() {
	stereoMeter->setValues(svc.currentVolumeLeft(), svc.currentVolumeRight());
}

void VControl::setVolume(int vol) {
	if(inUpdate) return;

	float volume = pow(2.0,(400.0 - vol)/200.0);
	if(vol == 1200) volume = 0.0;

	svc.scaleFactor(volume);
}

void VControl::updateVolume(float volume) {
	/*
	 * the inUpdate thing is important, because it stops us from sending the
	 * value to the server again - if we did that, two clients could bounce
	 * volume changes between the server and the clients
	 */
	inUpdate = true;
	slider->setValue( int( ( volume < 0.001 )?1200:( 400 - 200 * log( volume )/log(2.0) ) ) );
	inUpdate = false;
}

void VControl::useOldVolumeBar(int old) {
	delete stereoMeter;
	if (old) {
		stereoMeter= new StereoLedMeters(this);
	} else {
		stereoMeter= new PeakLevelMeters(this);
	}
	boxLayout->insertWidget(0, stereoMeter);
	stereoMeter->show();
}

void VControl::showScopeView() {
	if(!scopeView) {
		scopeView = new FFTScopeView(server);
		connect(scopeView,SIGNAL(closed()),this,SLOT(showScopeView()));
	} else {
		delete scopeView;
		scopeView = 0;
	}
}

void VControl::showFreeVerbView() {
	if(!freeVerbView) {
		freeVerbView = new FreeVerbView(server);
		connect(freeVerbView,SIGNAL(closed()),this,SLOT(showFreeVerbView()));
	} else {
		delete freeVerbView;
		freeVerbView = 0;
	}
}

void VControl::showEnvironment() {
	if(!environmentView) {
		environmentView = new EnvironmentView(defaultEnvironment());
		connect(environmentView,SIGNAL(closed()),this,SLOT(showEnvironment()));
	} else {
		delete environmentView;
		environmentView = 0;
	}
}

void VControl::showAudioManager() {
	if(!audioManager) {
		audioManager = new Gui_AUDIO_MANAGER();
		connect(audioManager,SIGNAL(closed()),this,SLOT(showAudioManager()));
	} else {
		delete audioManager;
		audioManager = 0;
	}
}

void VControl::showMediaView() {
	if(!mediaView) {
		mediaView = new MediaTypesView();
		connect(mediaView,SIGNAL(closed()),this,SLOT(showMediaView()));
	} else {
		delete mediaView;
		mediaView = 0;
	}
}

void VControl::showArtsStatusView() {
	if(!artsStatusView) {
		artsStatusView = new ArtsStatusView(server);
		connect(artsStatusView,SIGNAL(closed()),this,SLOT(showArtsStatusView()));
	} else {
		delete artsStatusView;
		artsStatusView = 0;
	}
}

void VControl::showMidiManager() {
	if(!midiManagerView) {
		midiManagerView = new MidiManagerView();
		connect(midiManagerView,SIGNAL(closed()),this,SLOT(showMidiManager()));
	} else {
		delete midiManagerView;
		midiManagerView = 0;
	}
}

void MainWindow::toggleVolumeBar() {
	vc->useOldVolumeBar(showOldVolumeDisplay->isChecked());
}

MainWindow::MainWindow() : KMainWindow(0) {
	vc = new VControl(this);
	setCentralWidget(vc);
	( void )new KAction( i18n("Toggle &FFT Scope"), 0, vc, SLOT( showScopeView() ), actionCollection(), "view_fft_scope" );
	( void )new KAction( i18n("Toggle &Audio Manager"), 0, vc, SLOT( showAudioManager() ), actionCollection(), "view_audio_manager" );
	( void )new KAction( i18n("Toggle aRts &Status"), 0, vc, SLOT( showArtsStatusView() ), actionCollection(), "view_arts_statusview" );
	( void )new KAction( i18n("Toggle &MIDI Manager"), 0, vc, SLOT( showMidiManager() ), actionCollection(), "view_midi_manager" );
	( void )new KAction( i18n("Toggle &Environment"), 0, vc, SLOT( showEnvironment() ), actionCollection(), "view_environment" );
	( void )new KAction( i18n("Toggle Available Media &Types"), 0, vc, SLOT( showMediaView()), actionCollection(), "view_media_types" );
	( void )new KAction( i18n("Toggle Free&Verb"), 0, vc, SLOT( showFreeVerbView() ), actionCollection(), "view_freeverb" );
	showOldVolumeDisplay=
		new KToggleAction( i18n("&LED-Style Volume Display"), 0, this,
		SLOT( toggleVolumeBar() ), actionCollection(), "old_volume_display" );
	( void ) KStdAction::quit( this, SLOT( close() ), actionCollection(), "quit_artscontrol" );

	createGUI("artscontrol.rc");
	resize(20,300);
	show();
}

int main(int argc, char **argv) {
	KAboutData aboutData( "artscontrol", I18N_NOOP("aRts control"),
		ARTS_VERSION, I18N_NOOP("Control tool for the aRts server"),
			KAboutData::License_GPL, "(c) 2000 Stefan Westerfeld");
	aboutData.addAuthor( "Stefan Westerfeld", I18N_NOOP( "Author and aRts maintainer" ), "stefan@space.twc.de" );
	aboutData.addCredit( "Arnold Krille", I18N_NOOP( "Some improvements" ), "arnold@arnoldarts.de" );

	KCmdLineArgs::init( argc, argv, &aboutData );

	KApplication app(argc, argv);

	// setup mcop communication
	QIOManager qiomanager;
	Dispatcher dispatcher(&qiomanager);

	ObjectManager::the()->provideCapability("kdegui");

	app.setMainWidget(new MainWindow);
	app.setName("artsbuilder");
	return app.exec();
}
