/*

    Copyright (C) 2003 Nikolas Zimmermann <wildfox@kde.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

*/

#include <fcntl.h>
#include <stdlib.h>

#include <kurl.h>
#include <kdebug.h>

#include <qfile.h>

#include "globals.h"
#include "wavewriter.h"

WaveWriter::WaveWriter() : ReaderWriter()
{
	m_numSamples = 0;
}

WaveWriter::~WaveWriter()
{
}

bool WaveWriter::finishFile()
{
	// 1. Flush any remaining data
	if(fwrite(m_buf, 1, m_pos, openStream()) != m_pos)
		m_eof = true;

	m_pos = 0;

	// 2. Free buffer
	free(m_buf);

	// 3. Seek back to start to fix the header
	if(fseek(openStream(), 0L, SEEK_SET) != 0)
	{
		kdDebug() << "Can't rewind output file to rewrite wav header!" << endl;
		return false;
	}

	// 4. Write last wave header!
	bool ok = writeWaveHeader(true);

	// 5. Close stream
	if(ok)
		fclose(openStream());
	
	return ok;
}

// Helper for writeWaveHeader
#define CHECK(test) if(!test) { kdDebug() << "Error while writing wave header!" << endl; return false; }

bool WaveWriter::writeWaveHeader(bool correctedHeader)
{
	// PCM-Compatible WAVE Header
	// bytes		variable			description
	// 0  - 3		'RIFF'
	// 4  - 7		wRiffLength			length of file minus the 8 byte riff header
	// 8  - 11		'WAVE'
	// 12 - 15		'fmt '
	// 16 - 19		wFmtSize			length of format chunk minus 8 byte header
	// 20 - 21		wFormatTag			identifies PCM, ULAW etc
	// 22 - 23		wChannels			number of channels
	// 24 - 27		dwSamplesPerSecond	samples per second per channel
	// 28 - 31		dwAvgBytesPerSec	non-trivial for compressed formats
	// 32 - 33		wBlockAlign			basic block size
	// 34 - 35		wBitsPerSample		non-trivial for compressed formats
	// 36 - 39		'data'
	// 40 - 43		dwDataLength		length of data chunk minus 8 byte header
	// 44 - (dwDataLength + 43)			the data

	Q_UINT32 wRiffLength = 0;
	Q_UINT16 wFmtSize = 16;					
	Q_UINT16 wFormatTag = 0x0001;
	Q_UINT16 wChannels = GLOBAL_CHANNELS;
	Q_UINT32 dwSamplesPerSecond = GLOBAL_RATE;
	Q_UINT16 wBitsPerSample = 16;
	Q_UINT16 wBlockAlign = wChannels * ((wBitsPerSample + 7) / 8);
	Q_UINT32 dwDataLength = 0x7ffff000L;
	Q_UINT16 wSamplesPerBlock = 1; // samples per channel per block
	Q_UINT32 dwAvgBytesPerSec = (Q_UINT32) ((double) wBlockAlign * GLOBAL_RATE / (double) wSamplesPerBlock + 0.5);
	
	// Fixup dwDataLength!
	// (m_numSamples was 0 on first writing!)
	if(correctedHeader)
	{
		// Long version
		// Q_UINT32 dwSamplesWritten = m_numSamples / wChannels;
		// Q_LONG blocksWritten = (dwSamplesWritten + wSamplesPerBlock - 1) / wSamplesPerBlock;
		// dwDataLength = blocksWritten * wBlockAlign;	

		dwDataLength = (m_numSamples / wChannels) * wBlockAlign; 
	}
	
	// Calculate riff length
	wRiffLength = 4 + (8 + wFmtSize) + (8 + dwDataLength);
	
	// Write header!
	CHECK(writeString("RIFF"));
	CHECK(write32Number(wRiffLength));
	CHECK(writeString("WAVE"));
	CHECK(writeString("fmt "));
	CHECK(write32Number(wFmtSize));
	CHECK(write16Number(wFormatTag));
	CHECK(write16Number(wChannels));
	CHECK(write32Number(dwSamplesPerSecond));
	CHECK(write32Number(dwAvgBytesPerSec));
	CHECK(write16Number(wBlockAlign));
	CHECK(write16Number(wBitsPerSample));
	CHECK(writeString("data"));
	CHECK(write32Number(dwDataLength));

	// All worked fine!
	return true;
}

Q_INT32 WaveWriter::writeWaveData(Q_INT32 *buf, Q_INT32 nsamp)
{
	Q_INT32 len, done = 0;

	while(done < nsamp)
	{
		if(m_pos > (m_size - m_signalSize))
		{
			if(fwrite(m_buf, 1, m_pos, openStream()) != m_pos)
				m_eof = true;
			
			m_pos = 0;
		}

		len = QMIN(nsamp - done, (Q_INT32) (m_size - m_pos) / m_signalSize);
		
		if(len)
		{
			writeBuffer(m_buf + m_pos, buf + done, len);
			m_pos += (len * m_signalSize);
			done += len;
		}
	}

	m_numSamples += done; // must later be divided by wChannels
	return done;
}

void WaveWriter::writeBuffer(char *buf1, Q_INT32 *buf2, Q_INT32 len)
{
	while(len)
	{
		Q_UINT16 datum;

		datum = ((Q_UINT16)((*buf2++) >> 16));
		
		if(GLOBAL_SWAP)
			datum = ((datum >> 8) | (datum << 8)) & 0xffff;
		
		*(Q_UINT16 *) buf1 = datum;
		buf1++; buf1++;
		len--;
	}
}
